CON
  _clkmode     = xtal1 + pll16x
  _xinfreq     = 5_000_000
  SPDIF_PIN    = 13
  BPM          = 30'120  ' Tempo
  ATTENUATION  = 4       ' Power of two
  SAMPLE_FP    = $17C6F  ' Sample rate, fixed point ($100000000 / 44100)

OBJ
  spdif : "spdifOut_mod"

VAR
  word buffer[4]
  long countPtr
  long rate[4]

PUB main | i, temp
  OUTA[15..12] := %0101         ' 12,14 : 217R to ground
  DIRA[15..12] := %0000         ' 13    : signal thru 560R
  count_addr := spdif.getCountAddr
  rate1_addr := @rate+ 0
  rate2_addr := @rate+ 4
  rate3_addr := @rate+ 8
  rate4_addr := @rate+12
  buffer1_addr := @buffer+0     'R
  buffer2_addr := @buffer+2     'L
  buffer3_addr := @buffer+6     'L
  buffer4_addr := @buffer+4     'R
  spdif.setBuffer(@buffer, 2)
  spdif.start(SPDIF_PIN)
  cognew(@synth, 0)
  repeat
    i := 0
    repeat while WORD[@track1+i]
      rate[0] := WORD[@track1+i] * SAMPLE_FP
      rate[1] := WORD[@track2+i] * SAMPLE_FP
      rate[2] := WORD[@track3+i] * SAMPLE_FP
      rate[3] := WORD[@track4+i] * SAMPLE_FP
      i += 2
      waitcnt(cnt + clkfreq*60/BPM)

DAT
        org 0

synth   rdlong   t1, count_addr            ' Wait for the sample count to change
        cmp      t1, spdif_count wz
  if_z  jmp      #synth
        mov      spdif_count, t1

        ' This is a sawtooth-wave synthesizer. "rate" determines the
        ' current tone frequency, and the high bits of "accumulator"
        ' are used to generate a signed 16-bit audio sample.
        add      accum1, rate1             ' Update sawtooth wave
        mov      t1, accum1                ' Chop off low bits...
        rdlong   rate1, rate1_addr         ' Load wave rate
        shr      t1, #(16 + ATTENUATION)   '   and decrease the output volume some
        sub      t1, midpoint              ' Convert unsigned to signed samples
        wrword   t1, buffer1_addr          ' Write the next sample now!

        add      accum2, rate2             ' Update sawtooth wave
        mov      t1, accum2                ' Chop off low bits...
        rdlong   rate2, rate2_addr         ' Load wave rate
        shr      t1, #(16 + ATTENUATION)   '   and decrease the output volume some
        sub      t1, midpoint              ' Convert unsigned to signed samples
        wrword   t1, buffer2_addr          ' Write the next sample now!

        add      accum3, rate3             ' Update sawtooth wave
        mov      t1, accum3                ' Chop off low bits...
        rdlong   rate3, rate3_addr         ' Load wave rate
        shr      t1, #(16 + ATTENUATION)   '   and decrease the output volume some
        sub      t1, midpoint              ' Convert unsigned to signed samples
        wrword   t1, buffer3_addr          ' Write the next sample now!

        add      accum4, rate4             ' Update sawtooth wave
        mov      t1, accum4                ' Chop off low bits...
        rdlong   rate4, rate4_addr         ' Load wave rate
        shr      t1, #(16 + ATTENUATION)   '   and decrease the output volume some
        sub      t1, midpoint              ' Convert unsigned to signed samples
        wrword   t1, buffer4_addr          ' Write the next sample now!

        jmp      #synth

count_addr    long  0
spdif_count   long  0

buffer1_addr  long  0
buffer2_addr  long  0
buffer3_addr  long  0
buffer4_addr  long  0
rate1_addr    long  0
rate2_addr    long  0
rate3_addr    long  0
rate4_addr    long  0
accum1        long  0
accum2        long  0
accum3        long  0
accum4        long  0
rate1         long  0
rate2         long  0
rate3         long  0
rate4         long  0

midpoint      long  $8000 >> ATTENUATION   ' Offset to sample midpoint
cFFFF         long  $FFFF
t1            res   1
t2            res   1
              fit

track1        word  523, 659, 784, 659, 784,1046, 784, 659, 0, 0
track2        word  523, 659, 784, 659, 784,1046, 784, 659, 0, 0
track3        word  523, 587, 659, 698, 784, 880, 987,1046, 0, 0
track4        word 1046, 987, 880, 784, 698, 659, 587, 523, 0, 0

'440 466 494  523 554 587 622 659 698 734 784 831 880 932 987  1046
